#include <stdint.h>
#include "caml/mlvalues.h"
#include "caml/alloc.h"
#include "caml/memory.h"
#include "caml/fail.h"

// Warning: this code is Intel-specific; it assumes little-endian arch and make unaligned memory accesses

CAMLprim value ocaml_pack_generic(value v_fmt, value v_rec)
{
    CAMLparam2(v_fmt, v_rec);
    CAMLlocal1(v_buf);
    int i, j;
    size_t sz;
    char *p;

    // calc needed buffer size

    sz = 0;

    for (i = 0; i < caml_string_length(v_fmt); i++) {
        switch (String_val(v_fmt)[i]) {
        case 'b': sz++; break;
        case 's':
        case 'S': sz += 2; break;
        case 'i': sz += 4; break;
        case 'l': sz += 8; break;
        default: goto end_of_fmt;
        }
    }
end_of_fmt:
    j = i;

    // alloc buffer

    v_buf = caml_alloc_string(sz);

    // serialize

    p = String_val(v_buf);

    for (i = 0; i < j; i++) {
        switch (String_val(v_fmt)[i]) {
        case 'b': *p++           = Int_val  (Field(v_rec, i));         break;
        case 's': *(uint16_t *)p = Int_val  (Field(v_rec, i)); p += 2; break;
        case 'S':
        {
            uint16_t u = Int_val(Field(v_rec, i));
            *p++ = u >> 8;
            *p++ = u & 255;
        }
        break;
        case 'i': *(uint32_t *)p = Int32_val(Field(v_rec, i)); p += 4; break;
        case 'l': *(uint64_t *)p = Int64_val(Field(v_rec, i)); p += 8; break;
        }
    }

    CAMLreturn(v_buf);
}

static uint16_t be16(uint8_t *p)
{
    return (*p << 8) + *(p+1);
}

CAMLprim value ocaml_unpack_generic(value v_fmt, value v_pos, value v_buf)
{
    CAMLparam3(v_fmt, v_pos, v_buf);
    CAMLlocal1(v_rec);
    int i, j = Int_val(v_pos), k = caml_string_length(v_buf), n = caml_string_length(v_fmt);

#define BUF(j) (String_val(v_buf)[j])

    v_rec = caml_alloc_tuple(n);

    for (i = 0; i < n; i++) {
        switch (String_val(v_fmt)[i]) {
        case 'b': if (j + 1 > k) goto raise; Field(v_rec, i) = Val_int(*(uint8_t  *) &BUF(j));         j++;    break;
        case 's': if (j + 2 > k) goto raise; Field(v_rec, i) = Val_int(*(uint16_t *) &BUF(j));         j += 2; break;
        case 'S': if (j + 2 > k) goto raise; Field(v_rec, i) = Val_int(be16((uint8_t *)&BUF(j)));      j += 2; break;
        case 'i': if (j + 4 > k) goto raise; Field(v_rec, i) = caml_copy_int32(*(uint32_t *) &BUF(j)); j += 4; break;
        case 'l': if (j + 8 > k) goto raise; Field(v_rec, i) = caml_copy_int64(*(uint64_t *) &BUF(j)); j += 8; break;
        default: goto end_of_fmt;
        }
    }
end_of_fmt:;

    CAMLreturn(v_rec);

raise:
    caml_invalid_argument("ocaml_unpack_generic: pos");
}

