#include <stdlib.h>
#include <poll.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/signals.h>

#define Nil Val_int(0)

CAMLprim value ocaml_poll(value v_lst, value v_timeout)
{
    CAMLparam2(v_lst, v_timeout);
    CAMLlocal3(v_res, v_nod, v_fd_ev);
    struct pollfd *fds;
    int i, n, n_ready, timeout;

    // List.length

    n = 0;
    for (v_nod = v_lst; v_nod != Nil; v_nod = Field(v_nod, 1))
        n++;

    if (0 == n) {
        CAMLreturn(Nil);
    }

    // alloc array of structs

    fds = calloc(sizeof(fds[0]), n);

    if (! fds) {
        CAMLreturn(Nil);
    }

    // convert list to array

    i = 0;
    for (v_nod = v_lst; v_nod != Nil; v_nod = Field(v_nod, 1)) {
        v_fd_ev = Field(v_nod, 0);
        fds[i].fd     = Int_val(Field(v_fd_ev, 0));
        fds[i].events = Int_val(Field(v_fd_ev, 1));
        i++;
    }

    // convert timeout from float

    {
        double d = Double_val(v_timeout);
        if (d < 0.) {
          timeout = -1; // blocking
        }
        else
          timeout = (int) (1e3 * d); // convert seconds to milliseconds
    }

    // call poll

    caml_enter_blocking_section();
    n_ready = poll(fds, n, timeout);
    caml_leave_blocking_section();

    // convert back to list

    v_res = Nil;

    for (i = 0; i < n; i++) {
        if (fds[i].revents) {
            v_fd_ev = caml_alloc_tuple(2);
            Field(v_fd_ev, 0) = Val_int(fds[i].fd);
            Field(v_fd_ev, 1) = Val_int(fds[i].revents);

            v_nod = caml_alloc_tuple(2);
            Field(v_nod, 0) = v_fd_ev;
            Field(v_nod, 1) = v_res;

            v_res = v_nod;
        }
    }

    // free array

    free(fds);

    CAMLreturn(v_res);
}


