#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include "rijndael.h"

#define KEYBITS 128

value ocaml_setup_encrypt(value v_key)
{
  CAMLparam1(v_key);
  CAMLlocal1(v_round_keys);

  // alloc round keys as a string on the OCaml heap

  v_round_keys = caml_alloc_string(4 * RKLENGTH(KEYBITS));

  rijndaelSetupEncrypt((void *) String_val(v_round_keys), (void *) String_val(v_key), 4, KEYBITS);

  CAMLreturn(v_round_keys);
}

value ocaml_setup_decrypt(value v_key)
{
  CAMLparam1(v_key);
  CAMLlocal1(v_round_keys);

  v_round_keys = caml_alloc_string(4 * RKLENGTH(KEYBITS));

  rijndaelSetupDecrypt((void *) String_val(v_round_keys), (void *) String_val(v_key), 4, KEYBITS);

  CAMLreturn(v_round_keys);
}

value ocaml_encrypt_block(value v_round_keys, value v_blk)
{
  CAMLparam2(v_round_keys, v_blk);
  CAMLlocal1(v_enc);

  v_enc = caml_alloc_string(16);

  rijndaelEncrypt((void *) String_val(v_round_keys), 4, (void *) String_val(v_blk), (void *) String_val(v_enc));

  CAMLreturn(v_enc);
}

value ocaml_decrypt_block(value v_round_keys, value v_blk)
{
  CAMLparam2(v_round_keys, v_blk);
  CAMLlocal1(v_dec);

  v_dec = caml_alloc_string(16);

  rijndaelDecrypt((void *) String_val(v_round_keys), 4, (void *) String_val(v_blk), (void *) String_val(v_dec));

  CAMLreturn(v_dec);
}

value ocaml_xor_blocks(value v_a, value v_b)
{
  CAMLparam2(v_a, v_b);
  CAMLlocal1(v_x);
  int la = caml_string_length(v_a);
  int lb = caml_string_length(v_b);
  int lx = (la < lb) ? la : lb;
  int i;
  char *pa;
  char *pb;
  char *px;

  v_x = caml_alloc_string(lx);

  pa = String_val(v_a);
  pb = String_val(v_b);
  px = String_val(v_x);

  for (i = 0; i < lx; i++) {
    *px++ = *pa++ ^ *pb++;
  }

  CAMLreturn(v_x);
}

