
type enc_key
type dec_key

external setup_encrypt : key:bytes -> enc_key = "ocaml_setup_encrypt"
external setup_decrypt : key:bytes -> dec_key = "ocaml_setup_decrypt"
external encrypt_block : enc_key -> bytes -> bytes = "ocaml_encrypt_block"
external decrypt_block : dec_key -> bytes -> bytes = "ocaml_decrypt_block"
external xor_blocks : bytes -> bytes -> bytes = "ocaml_xor_blocks" (* to facilitate CBC *)

let cbc_encrypt enc_key buf =
  let iv = Bytes.make 16 '\x00' in (* always use a nul IV *)
  let todo = ref (Bytes.length buf) in
  let n = (!todo + 15) / 16 + 1 in
  let enc = Bytes.create (16 * n) in
  let partial = Bytes.create 16 in
  Bytes.blit iv 0 enc 0 16;
  for i = 1 to n - 1 do
    Bytes.blit buf ((i - 1) * 16) partial 0 (min 16 !todo);
    let blk = xor_blocks partial iv in
    let enc_blk = encrypt_block enc_key blk in
    Bytes.blit enc_blk 0 enc (i * 16) 16;
    Bytes.blit enc_blk 0 iv 0 16;
    todo := !todo - 16
  done;
  enc

let cbc_decrypt dec_key buf =
  let n = Bytes.length buf / 16 - 1 in
  let dec = Bytes.create (16 * n) in
  for i = 1 to n do
    let iv = Bytes.sub buf ((i - 1) * 16) 16 in
    let plain = xor_blocks iv (decrypt_block dec_key (Bytes.sub buf (i * 16) 16)) in
    Bytes.blit plain 0 dec ((i - 1) * 16) 16
  done;
  dec

